(function() {
  var AbstractPropertyManager, createObject, deepEqual, defineProperty, getkeys, isArray, isBoolean, isFunction, isObject, isString,
    indexOf = [].indexOf || function(item) { for (var i = 0, l = this.length; i < l; i++) { if (i in this && this[i] === item) return i; } return -1; };

  defineProperty = require('util-ex/lib/defineProperty');

  isFunction = require('util-ex/lib/is/type/function');

  isObject = require('util-ex/lib/is/type/object');

  isString = require('util-ex/lib/is/type/string');

  isBoolean = require('util-ex/lib/is/type/boolean');

  isArray = require('util-ex/lib/is/type/array');

  createObject = require('inherits-ex/lib/createObject');

  deepEqual = require('deep-equal');

  getkeys = Object.keys;

  module.exports = AbstractPropertyManager = (function() {
    function AbstractPropertyManager() {}

    AbstractPropertyManager.prototype.nonExported1stChar = '$';

    AbstractPropertyManager.prototype.constrcutor = function() {
      return this.initialize.apply(this, arguments);
    };

    AbstractPropertyManager.prototype.assignPropertyTo = function(dest, src, name, value, attrs, options) {};

    AbstractPropertyManager.prototype.getProperties = function() {};

    AbstractPropertyManager.prototype.defineProperties = function(aProperties) {};

    AbstractPropertyManager.prototype.initialize = function(src) {
      if (src == null) {
        src = {};
      }
      this.defineProperties(src.attributes);
      return this.assign(src);
    };

    AbstractPropertyManager.prototype.clone = function(options) {
      return this.cloneTo({}, options);
    };

    AbstractPropertyManager.prototype.cloneTo = function(dest, options) {
      var result;
      dest = this.mergeTo(dest, options);
      result = createObject(this.Class || this.constructor);
      return result.assign(dest);
    };

    AbstractPropertyManager.prototype.assign = function(src, aOptions) {
      return this.__assign(src, aOptions);
    };

    AbstractPropertyManager.prototype.__assign = function(src, aOptions) {
      var defaultOptions, exclude, i, k, len, n, overwrite, v, vAlias, vAttrs, vName;
      defaultOptions = this.defaultOptions;
      if (!aOptions) {
        aOptions = (defaultOptions && defaultOptions.assign) || {};
      }
      if (aOptions.overwrite == null) {
        aOptions.overwrite = true;
      }
      aOptions.isExported = false;
      exclude = aOptions.exclude, overwrite = aOptions.overwrite;
      if (isString(exclude)) {
        exclude = [exclude];
      } else if (!isArray(exclude)) {
        exclude = [];
      }
      vAttrs = this.getProperties();
      for (k in vAttrs) {
        v = vAttrs[k];
        if (indexOf.call(exclude, k) >= 0) {
          continue;
        }
        if (src.hasOwnProperty(k)) {
          vName = k;
        } else if ((vName = v && v.name) && (!src.hasOwnProperty(vName) || (indexOf.call(exclude, vName) >= 0))) {
          vName = null;
        } else if (!vName && (vAlias = v && v.alias)) {
          if (isString(vAlias)) {
            if (src.hasOwnProperty(vAlias) && !(indexOf.call(exclude, vAlias) >= 0)) {
              vName = vAlias;
            }
          } else if (isArray(vAlias)) {
            for (i = 0, len = vAlias.length; i < len; i++) {
              n = vAlias[i];
              if (src.hasOwnProperty(n) && !(indexOf.call(exclude, n) >= 0)) {
                vName = n;
                break;
              }
            }
          }
        }
        if (!vName) {
          continue;
        }
        v = src[vName];
        if (overwrite || this[k] === void 0) {
          this.assignPropertyTo(this, src, k, v, vAttrs, aOptions);
        }
      }
      if (isFunction(this._assign)) {
        this._assign(src, aOptions);
      }
      return this;
    };

    AbstractPropertyManager.prototype.assignProperty = function(src, name, value, attrs, options) {
      var defaultOptions;
      defaultOptions = this.defaultOptions;
      options = options || (defaultOptions && defaultOptions.assign) || {};
      if (options.overwrite == null) {
        options.overwrite = true;
      }
      options.isExported = false;
      this.assignPropertyTo(this, src, name, value, attrs, options);
    };

    AbstractPropertyManager.prototype.mergeTo = function(dest, aOptions) {
      var exclude, isExported, k, overwrite, ref, skipDefault, skipNull, skipReadOnly, skipUndefined, v, vAttrs, vIsReadonly, val;
      if (aOptions == null) {
        aOptions = {};
      }
      overwrite = aOptions.overwrite, exclude = aOptions.exclude, skipDefault = aOptions.skipDefault, skipReadOnly = aOptions.skipReadOnly, isExported = aOptions.isExported, skipNull = aOptions.skipNull, skipUndefined = aOptions.skipUndefined;
      if (isString(exclude)) {
        exclude = [exclude];
      } else if (!isArray(exclude)) {
        exclude = [];
      }
      if (dest == null) {
        dest = {};
      }
      vAttrs = this.getProperties();
      for (k in vAttrs) {
        v = vAttrs[k];
        if (indexOf.call(exclude, k) >= 0) {
          continue;
        }
        if (v && v.name && (ref = v.name, indexOf.call(exclude, ref) >= 0)) {
          continue;
        }
        if (skipNull && this[k] === null) {
          continue;
        }
        if (skipUndefined && this[k] === void 0) {
          continue;
        }
        vIsReadonly = v.writable === false || (v.get && !v.set);
        if (skipReadOnly && v.exported !== true && vIsReadonly) {
          continue;
        }
        if (overwrite || dest[k] === void 0) {
          val = aOptions[k];
          if (val == null) {
            val = this[k];
          }
          this.assignPropertyTo(dest, this, k, val, vAttrs, aOptions);
        }
      }
      return dest;
    };

    AbstractPropertyManager.prototype.exportTo = function(dest, aOptions) {
      var defaultOptions;
      defaultOptions = this.defaultOptions;
      if (!aOptions) {
        aOptions = (defaultOptions && defaultOptions["export"]) || {};
      }
      if (aOptions.skipDefault == null) {
        aOptions.skipDefault = true;
      }
      if (aOptions.skipUndefined == null) {
        aOptions.skipUndefined = true;
      }
      aOptions.isExported = true;
      return this.mergeTo(dest, aOptions);
    };

    AbstractPropertyManager.prototype.toObject = function(options) {
      var result;
      result = {};
      return this.exportTo(result, options);
    };

    AbstractPropertyManager.prototype.toJSON = function() {
      return this.toObject();
    };

    AbstractPropertyManager.prototype.assignTo = function(dest, aOptions) {
      var defaultOptions;
      defaultOptions = this.defaultOptions;
      if (!aOptions) {
        aOptions = (defaultOptions && defaultOptions.assign) || {};
      }
      aOptions.isExported = false;
      aOptions.overwrite = true;
      dest = this.mergeTo(dest, aOptions);
      if (isFunction(dest._assign)) {
        dest._assign(this, aOptions);
      }
      return dest;
    };

    AbstractPropertyManager.prototype.isSame = function(src, aOptions) {
      var k, ref, v;
      ref = this.assignTo({}, aOptions);
      for (k in ref) {
        v = ref[k];
        if (!deepEqual(src[k], v)) {
          return false;
        }
      }
      return true;
    };

    return AbstractPropertyManager;

  })();

  module.exports["default"] = module.exports;

}).call(this);

//# sourceMappingURL=abstract.js.map
