(function() {
  var NormalPropertyManager, PropertyManager, assignValue, cloneObject, deepEqual, defineProperties, defineProperty, extend, getAllOwnKeys, getOwnPropertyDescriptor, getPrototypeOf, getkeys, isArray, isFunction, isObject, isString, setPrototypeOf;

  defineProperty = require('util-ex/lib/defineProperty');

  isFunction = require('util-ex/lib/is/type/function');

  isObject = require('util-ex/lib/is/type/object');

  isString = require('util-ex/lib/is/type/string');

  isArray = require('util-ex/lib/is/type/array');

  extend = require('util-ex/lib/_extend');

  cloneObject = require('util-ex/lib/clone-object');

  getPrototypeOf = require('inherits-ex/lib/getPrototypeOf');

  setPrototypeOf = require('inherits-ex/lib/setPrototypeOf');

  deepEqual = require('deep-equal');

  assignValue = require('./assign-value');

  PropertyManager = require('./abstract');

  getkeys = Object.keys;

  getAllOwnKeys = Object.getOwnPropertyNames;

  defineProperties = Object.defineProperties;

  getOwnPropertyDescriptor = Object.getOwnPropertyDescriptor;

  module.exports = NormalPropertyManager = (function() {
    var copyProperties, defineObjectProperties, extendsObj, getRealAttrName;

    NormalPropertyManager.getProperties = function() {
      var k, ref, result, v;
      result = {};
      if (this.prototype.$attributes) {
        ref = this.prototype.$attributes;
        for (k in ref) {
          v = ref[k];
          result[k] = v;
        }
      }
      return result;
    };

    getRealAttrName = function(attrs, name) {
      var k, v;
      if (!attrs[name]) {
        for (k in attrs) {
          v = attrs[k];
          if (v.name === name) {
            return k;
          }
        }
        return;
      }
      return name;
    };

    function NormalPropertyManager() {
      this.initialize.apply(this, arguments);
    }

    extend(NormalPropertyManager.prototype, PropertyManager.prototype);

    NormalPropertyManager.prototype.$attributes = null;

    defineProperty(NormalPropertyManager, '$attributes', void 0, {
      get: function() {
        return getPrototypeOf(this).$attributes;
      }
    });

    NormalPropertyManager.prototype.getProperties = function() {
      return this.$attributes;
    };

    copyProperties = function(aProperties, dest, nonExported1stChar) {
      var k, v, vWritable;
      if (dest == null) {
        dest = {};
      }
      if (aProperties) {
        for (k in aProperties) {
          v = aProperties[k];
          if (!(!isArray(v) && isObject(v))) {
            v = {
              value: v
            };
          }
          if ((v.enumerable == null) && v.assigned === false && v.exported === false) {
            v.enumerable = false;
          } else {
            v.enumerable = v.enumerable !== false;
          }
          vWritable = v.writable !== false || isFunction(v.set);
          if (v.assigned == null) {
            v.assigned = v.enumerable && vWritable;
          }
          if (v.exported == null) {
            v.exported = v.enumerable && k[0] !== nonExported1stChar && vWritable;
          }
          dest[k] = v;
        }
      }
      return dest;
    };

    extendsObj = function(obj) {
      var result;
      result = {};
      setPrototypeOf(result, obj);
      return result;
    };

    NormalPropertyManager.defineProperties = defineObjectProperties = function(aTarget, aProperties, recreate) {
      var nonExported1stChar, vAttrs, vHasOwnProperty, vIsProperties, vPrototype;
      if (isFunction(aTarget)) {
        vPrototype = aTarget.prototype;
        nonExported1stChar = vPrototype.nonExported1stChar;
      } else if (isObject(aTarget)) {
        vPrototype = getPrototypeOf(aTarget);
        nonExported1stChar = aTarget.nonExported1stChar;
      } else {
        throw new TypeError('the target should be a ctor or object!');
      }
      if (nonExported1stChar == null) {
        nonExported1stChar = NormalPropertyManager.prototype.nonExported1stChar;
      }
      vAttrs = vPrototype.$attributes;
      vHasOwnProperty = vPrototype.hasOwnProperty('$attributes');
      vIsProperties = isObject(vAttrs);
      if (recreate !== true && !vHasOwnProperty && vIsProperties) {
        vPrototype.$attributes = vAttrs = extendsObj(vAttrs);
      } else if (recreate || !vIsProperties) {
        vAttrs = vPrototype.$attributes = {};
      }
      return copyProperties(aProperties, vAttrs, nonExported1stChar);
    };

    NormalPropertyManager.prototype.defineProperties = function(aProperties, recreate) {
      var k, v, vAttrs, value;
      if (recreate == null) {
        recreate = false;
      }
      vAttrs = defineObjectProperties(this, aProperties, recreate);
      for (k in vAttrs) {
        v = vAttrs[k];
        value = v.value;
        if (!v.get && !v.set && v.clone !== false && isObject(value)) {
          value = cloneObject(value);
        }
        value = assignValue(value, v.type);
        defineProperty(this, k, value, v);
      }
    };

    NormalPropertyManager.prototype.assignPropertyTo = function(dest, src, name, value, attrs, options) {
      var isExported, skipDefault, skipExists, vAttr, vCanAssign, vIsReadonly;
      if (options) {
        skipDefault = options.skipDefault, isExported = options.isExported, skipExists = options.skipExists;
      }
      if (!attrs) {
        attrs = this.getProperties();
      }
      name = getRealAttrName(attrs, name);
      if (name) {
        vAttr = attrs[name];
        if (skipExists && dest[name] !== void 0) {
          return;
        }
        if (!((vAttr.assigned && !isExported) || (vAttr.exported && isExported))) {
          return;
        }
        vIsReadonly = vAttr.writable === false || (vAttr.get && !vAttr.set);
        if (vIsReadonly && vAttr.exported !== true) {
          return;
        }
        if (skipDefault && !vIsReadonly && deepEqual(vAttr.value, value)) {
          return;
        }
        vCanAssign = (!isExported && vAttr.assigned) || value !== void 0;
        if (isFunction(vAttr.assign)) {
          value = vAttr.assign(value, dest, src, name, options);
        }
        if (isExported) {
          name = vAttr.name || name;
        }
        if (value === void 0 && vAttr.value !== void 0) {
          value = vAttr.value;
        }
        if (vCanAssign) {
          if (isExported) {
            if (value != null) {
              if (isFunction(value.toObject)) {
                value = value.toObject(options);
              } else if (isFunction(value.toJSON)) {
                value = value.toJSON();
              }
            }
            dest[name] = value;
          } else {
            dest[name] = assignValue(value, vAttr.type);
          }
        }
      }
    };

    return NormalPropertyManager;

  })();

  module.exports["default"] = module.exports;

}).call(this);

//# sourceMappingURL=normal.js.map
